/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2019 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.artifacts.browse.treebrowser.actions;

import org.apache.http.HttpStatus;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.common.ConstantValues;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.rest.common.model.RestModel;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.signature.SignedUrlService;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.action.DownloadArtifact;
import org.artifactory.util.HttpUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.servlet.http.HttpServletResponse;

/**
 * Returns a signed url for the current user authenticated per the path requested.
 *
 * NOTE: this component is undocumented, and should be restricted in the future - should not be available for all users!
 *
 * @author Yuval Reches
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class DownloadArtifactWithSignedUrlService implements RestService<DownloadArtifact> {
    private static final Logger log = LoggerFactory.getLogger(DownloadArtifactWithSignedUrlService.class);

    @Autowired
    private AuthorizationService authorizationService;

    @Autowired
    SignedUrlService signedUrlService;

    @Override
    public void execute(ArtifactoryRestRequest<DownloadArtifact> request, RestResponse response) {
        DownloadArtifact downloadArtifact = request.getImodel();
        RepoPath repoPath = RepoPathFactory.create(downloadArtifact.getRepoKey(), downloadArtifact.getPath());

        String errorMsg;
        if (!authorizationService.canRead(repoPath)) {
            errorMsg = "Forbidden UI REST call from user: '" + authorizationService.currentUsername() + "'";
            response.responseCode(HttpServletResponse.SC_FORBIDDEN).error(errorMsg).buildResponse();
            log.error(errorMsg);
            return;
        }

        String contextUrl = HttpUtils.getServletContextUrl(request.getServletRequest());
        RestModel artifactDownloadModel = getArtifactDownloadModel(contextUrl, repoPath);
        response.responseCode(HttpStatus.SC_CREATED);
        response.iModel(artifactDownloadModel);
    }

    /**
     * Build signed url and update download model
     */
    private RestModel getArtifactDownloadModel(String contextUrl, RepoPath repoPath) {
        DownloadArtifact downloadArtifact = new DownloadArtifact();
        String downloadPath = signedUrlService.createSignedUrl(contextUrl, repoPath.toPath(),
                ConstantValues.signedUrlUiDownloadValidForSeconds.getLong(), true);
        downloadArtifact.setDownloadPath(downloadPath);
        return downloadArtifact;
    }

}
