package org.artifactory.ui.rest.service.builds.buildsinfo.tabs.permissions;

import org.apache.http.HttpStatus;
import org.artifactory.api.build.BuildService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.api.security.UserGroupService;
import org.artifactory.rest.common.model.FeedbackMsg;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.ArtifactoryRestResponse;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.util.JsonUtil;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.io.IOException;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;

/**
 * @author Yevdo Abramov
 * Created on 27/10/2020
 */
public class GetBuildEffectivePermissionsByEntityServiceTest {

    @Mock
    private AuthorizationService authService;
    @Mock
    private UserGroupService userGroupService;
    @Mock
    private BuildService buildService;

    private GetBuildEffectivePermissionsByEntityService service;

    @BeforeMethod
    public void setUp() {
        MockitoAnnotations.initMocks(this);
        service = new GetBuildEffectivePermissionsByEntityService(authService, userGroupService, buildService);
    }

    @Test
    public void testExecuteWithInvalidQueryParamsSpecified() throws IOException {
        ArtifactoryRestRequest request = mock(ArtifactoryRestRequest.class);
        RestResponse response = new ArtifactoryRestResponse();
        when(buildService.normaliseBuildRepoName(any())).thenReturn("buildRepo");

        service.execute(request, response);
        assertEquals(response.getResponseCode(), HttpStatus.SC_BAD_REQUEST);
        FeedbackMsg error = JsonUtil.mapDataToModel((String) response.getEntity(), FeedbackMsg.class);
        assertEquals(error.getError(),
                "One or more of the following parameters is missing: buildName, buildNumber, buildDate");

        when(request.getQueryParamByKey("buildName")).thenReturn("my_build");
        service.execute(request, response);
        assertEquals(response.getResponseCode(), HttpStatus.SC_BAD_REQUEST);
        error = JsonUtil.mapDataToModel((String) response.getEntity(), FeedbackMsg.class);
        assertEquals(error.getError(),
                "One or more of the following parameters is missing: buildName, buildNumber, buildDate");

        when(request.getQueryParamByKey("buildNumber")).thenReturn("1");
        service.execute(request, response);
        assertEquals(response.getResponseCode(), HttpStatus.SC_BAD_REQUEST);
        error = JsonUtil.mapDataToModel((String) response.getEntity(), FeedbackMsg.class);
        assertEquals(error.getError(),
                "One or more of the following parameters is missing: buildName, buildNumber, buildDate");

        when(request.getQueryParamByKey("buildDate")).thenReturn("ds1");
        service.execute(request, response);
        assertEquals(response.getResponseCode(), HttpStatus.SC_BAD_REQUEST);
        error = JsonUtil.mapDataToModel((String) response.getEntity(), FeedbackMsg.class);
        assertEquals(error.getError(), "Invalid buildDate format");
        verify(request, times(4)).getQueryParamByKey("buildRepo");
    }

    @Test
    public void testExecuteWhenHasManagePermissions() {
        ArtifactoryRestRequest request = mock(ArtifactoryRestRequest.class);
        RestResponse response = new ArtifactoryRestResponse();
        when(buildService.normaliseBuildRepoName(any())).thenReturn("buildRepo");
        when(request.getQueryParamByKey("buildName")).thenReturn("my_build");
        when(request.getQueryParamByKey("buildNumber")).thenReturn("1");
        when(request.getQueryParamByKey("buildDate")).thenReturn("100");
        when(authService.canManageBuild(anyString(), anyString(), anyString(), anyString())).thenReturn(true);

        service.execute(request, response);

        verify(authService).canManageBuild(anyString(), anyString(), anyString(), anyString());
        verify(userGroupService).getBuildPathAcls(anyString(), anyString(), anyString(), any());
        verify(buildService).normaliseBuildRepoName(any());
        verify(request).getQueryParamByKey("buildRepo");
        verifyNoMoreInteractions(authService);
        verifyNoMoreInteractions(userGroupService);
        verifyNoMoreInteractions(buildService);
    }

    @Test
    public void testExecuteWhenNoPermissions() {
        ArtifactoryRestRequest request = mock(ArtifactoryRestRequest.class);
        RestResponse response = new ArtifactoryRestResponse();
        when(buildService.normaliseBuildRepoName(any())).thenReturn("buildRepo");
        when(request.getQueryParamByKey("buildName")).thenReturn("my_build");
        when(request.getQueryParamByKey("buildNumber")).thenReturn("1");
        when(request.getQueryParamByKey("buildDate")).thenReturn("100");
        when(authService.canManageBuild(anyString(), anyString(), anyString(), anyString())).thenReturn(false);

        service.execute(request, response);

        verify(authService).canManageBuild(anyString(), anyString(), anyString(), anyString());
        verify(buildService).normaliseBuildRepoName(any());
        verify(request).getQueryParamByKey("buildRepo");
        verifyNoMoreInteractions(authService);
        verifyNoMoreInteractions(userGroupService);
        verifyNoMoreInteractions(buildService);

        assertEquals(response.getResponseCode(), HttpStatus.SC_FORBIDDEN);
    }
}