package org.artifactory.ui.rest.service.onboarding;

import com.google.common.collect.Lists;
import org.artifactory.addon.ArtifactoryRunningMode;
import org.artifactory.addon.OssAddonsManager;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.ui.rest.model.onboarding.OnboardingRepoState;
import org.testng.Assert;
import org.testng.annotations.Test;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static org.artifactory.ui.rest.model.onboarding.OnboardingRepoState.ALREADY_SET;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.when;

/**
 * @author dudim
 */
public class GetUnsetReposServiceTest {

    @Test
    public void testGetSupportedRepoTypesMapJcrTypeWithoutConan() {
        OssAddonsManager addonsManagerSpy = spy(new OssAddonsManager());
        CentralConfigService configService = mock(CentralConfigService.class);
        MutableCentralConfigDescriptor descriptor = mock(MutableCentralConfigDescriptor.class);
        when(descriptor.getUrlBase()).thenReturn(null);
        when(configService.getMutableDescriptor()).thenReturn(descriptor);
        GetUnsetReposService getUnsetReposService = new GetUnsetReposService(null, addonsManagerSpy, configService);
        doReturn(ArtifactoryRunningMode.JCR).when(addonsManagerSpy).getArtifactoryRunningMode();
        Map<RepoType, OnboardingRepoState> supportedRepoTypesMap = getUnsetReposService.getSupportedRepoTypesMap();
        Assert.assertEquals(supportedRepoTypesMap.get(RepoType.Generic), OnboardingRepoState.UNSET);
        Assert.assertEquals(supportedRepoTypesMap.get(RepoType.Docker), OnboardingRepoState.UNSET);
        Assert.assertEquals(supportedRepoTypesMap.get(RepoType.Helm), OnboardingRepoState.UNSET);
        Assert.assertFalse(supportedRepoTypesMap.containsKey(RepoType.Cargo));
        List<RepoType> repoTypesList = new ArrayList<>(Arrays.asList(RepoType.values()));
        repoTypesList.removeAll(Lists.newArrayList(RepoType.Helm, RepoType.Docker, RepoType.Generic));
        repoTypesList.forEach(repoType -> Assert.assertNotEquals(repoType, OnboardingRepoState.UNSET));
    }

    @Test
    public void testGetSupportedRepoTypesMapJcrTypeWithConan() {
        OssAddonsManager addonsManagerSpy = spy(new OssAddonsManager());
        CentralConfigService configService = mock(CentralConfigService.class);
        MutableCentralConfigDescriptor descriptor = mock(MutableCentralConfigDescriptor.class);
        when(descriptor.getUrlBase()).thenReturn("http://localhost:8082");
        when(configService.getMutableDescriptor()).thenReturn(descriptor);
        GetUnsetReposService getUnsetReposService = new GetUnsetReposService(null, addonsManagerSpy, configService);
        doReturn(ArtifactoryRunningMode.JCR).when(addonsManagerSpy).getArtifactoryRunningMode();
        Map<RepoType, OnboardingRepoState> supportedRepoTypesMap = getUnsetReposService.getSupportedRepoTypesMap();
        Assert.assertEquals(supportedRepoTypesMap.get(RepoType.Generic), OnboardingRepoState.UNSET);
        Assert.assertEquals(supportedRepoTypesMap.get(RepoType.Docker), OnboardingRepoState.UNSET);
        Assert.assertEquals(supportedRepoTypesMap.get(RepoType.Helm), OnboardingRepoState.UNSET);
        Assert.assertTrue(supportedRepoTypesMap.containsKey(RepoType.Cargo));
        List<RepoType> repoTypesList = new ArrayList<>(Arrays.asList(RepoType.values()));
        repoTypesList.removeAll(Lists.newArrayList(RepoType.Helm, RepoType.Docker, RepoType.Generic));
        repoTypesList.forEach(repoType -> Assert.assertNotEquals(repoType, OnboardingRepoState.UNSET));
    }

    @Test
    public void testGetSupportedRepoTypesMapConanType() {
        OssAddonsManager addonsManagerSpy = spy(new OssAddonsManager());
        CentralConfigService configService = mock(CentralConfigService.class);
        MutableCentralConfigDescriptor descriptor = mock(MutableCentralConfigDescriptor.class);
        when(configService.getMutableDescriptor()).thenReturn(descriptor);
        GetUnsetReposService getUnsetReposService = new GetUnsetReposService(null, addonsManagerSpy, configService);
        doReturn(ArtifactoryRunningMode.CONAN).when(addonsManagerSpy).getArtifactoryRunningMode();
        Map<RepoType, OnboardingRepoState> supportedRepoTypesMap = getUnsetReposService.getSupportedRepoTypesMap();
        Assert.assertEquals(supportedRepoTypesMap.get(RepoType.Conan), OnboardingRepoState.UNSET);
        List<RepoType> repoTypesList = new ArrayList<>(Arrays.asList(RepoType.values()));
        repoTypesList.removeAll(Lists.newArrayList(RepoType.Conan));
        repoTypesList.forEach(repoType -> Assert.assertNotEquals(repoType, OnboardingRepoState.UNSET));
    }

    @Test
    public void testFilterRepoTypesMap() {
        RepositoryService repositoryService = mock(RepositoryService.class);
        when(repositoryService.getLocalRepoDescriptors()).thenReturn(List.of(createLocalRepo()));
        when(repositoryService.getFederatedRepoDescriptors()).thenReturn(List.of(createFedRepo()));
        Map<RepoType, OnboardingRepoState> repoStateMap = new HashMap<>();
        repoStateMap.put(RepoType.Docker, OnboardingRepoState.UNAVAILABLE);
        repoStateMap.put(RepoType.Npm, OnboardingRepoState.UNAVAILABLE);
        GetUnsetReposService unsetReposService = new GetUnsetReposService(repositoryService, null, null);
        unsetReposService.filterRepoTypesMap(repoStateMap);
        Assert.assertEquals(repoStateMap.get(RepoType.Docker), ALREADY_SET);
        Assert.assertEquals(repoStateMap.get(RepoType.Npm), ALREADY_SET);
    }

    private LocalRepoDescriptor createLocalRepo() {
        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setKey("local");
        localRepoDescriptor.setType(RepoType.Docker);
        return localRepoDescriptor;
    }

    private FederatedRepoDescriptor createFedRepo() {
        FederatedRepoDescriptor localRepoDescriptor = new FederatedRepoDescriptor();
        localRepoDescriptor.setKey("fed");
        localRepoDescriptor.setType(RepoType.Npm);
        return localRepoDescriptor;
    }
}